/*
 * Includes
 */
#include <linux/kernel.h>	/* printk() */
#include <linux/module.h>	/* modules */
#include <linux/init.h>		/* module_{init,exit}() */
#include <linux/fs.h>           /* file_operations */
#include <asm/uaccess.h>	/* copy_{from,to}_user() */
#include <linux/slab.h>		/* kmalloc()/kfree() */
#include <linux/miscdevice.h>   /* misc driver interface */

MODULE_DESCRIPTION("mydriver4");
MODULE_AUTHOR("Stelian Pop/Pierre Ficheux, Open Wide");
MODULE_LICENSE("GPL");

/*
 * Arguments
 */

static size_t buf_size = 64; /* Buffer size */
module_param(buf_size, int, 0644);
MODULE_PARM_DESC(buf_size, "Buffer size");

/*
 * Global variables
 */

static char *buffer;             /* copy_from/to_user buffer */
static struct miscdevice mymisc; /* Misc device handler */
static size_t num = 0; /* Number of available bytes in the buffer */

/*
 * File operations
 */
static ssize_t mydriver4_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
  size_t real;

  real = min(num, count);

  if (real)
    if (copy_to_user(buf, buffer, real))
      return -EFAULT;

  num = 0; /* Destructive read (no more data after a read) */

  printk(KERN_DEBUG "mydriver4: read %d/%d chars %s\n", real, count, buffer);

  return real;
}

static ssize_t mydriver4_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
  size_t real;

  real = min(buf_size, count);

  if (real)
    if (copy_from_user(buffer, buf, real))
      return -EFAULT;

  num = real; /* Destructive write (overwrite previous data if any) */

  printk(KERN_DEBUG "mydriver4: wrote %d/%d chars %s\n", real, count, buffer);

  return real;
}

static int mydriver4_open(struct inode *inode, struct file *file)
{
  printk(KERN_INFO "mydriver4: open()\n");

  return 0;
}

static int mydriver4_release(struct inode *inode, struct file *file)
{
  printk(KERN_INFO "mydriver4: release()\n");

  return 0;
}

static struct file_operations mydriver4_fops = {
  .owner   =	THIS_MODULE,
  .read    =	mydriver4_read,
  .write   =	mydriver4_write,
  .open    =	mydriver4_open,
  .release =	mydriver4_release,
};

/*
 * Init and Exit
 */
static int __init mydriver4_init(void)
{
  int ret;

  mymisc.minor = MISC_DYNAMIC_MINOR;
  mymisc.name = "mydriver4";
  mymisc.fops = &mydriver4_fops;

  ret = misc_register(&mymisc);

  if (ret < 0) {
    printk(KERN_WARNING "mydriver4: unable to get a dynamic minor\n");
    return ret;
  }

  buffer = (char *)kmalloc(buf_size, GFP_KERNEL);
  if (buffer != NULL) {
    printk(KERN_DEBUG "mydriver4: allocated a %d bytes buffer\n", buf_size);
  } else {
    printk(KERN_WARNING "mydriver4: unable to allocate a %d bytes buffer\n", buf_size);
    misc_deregister(&mymisc);

    return -ENOMEM;
  }

  return 0;
}

static void __exit mydriver4_exit(void)
{
  kfree (buffer);

  misc_deregister(&mymisc);

  printk(KERN_INFO "mydriver4: successfully unloaded\n");
}

/*
 * Module entry points
 */
module_init(mydriver4_init);
module_exit(mydriver4_exit);
