/*
 * Includes
 */
#include <linux/kernel.h>	/* printk() */
#include <linux/module.h>	/* modules */
#include <linux/init.h>		/* module_{init,exit}() */
#include <linux/slab.h>		/* kmalloc()/kfree() */
#include <linux/pci.h>		/* pci_*() */
#include <linux/pci_ids.h>	/* pci idents */
#include <linux/list.h>		/* list_*() */
#include <asm/uaccess.h>	/* copy_{from,to}_user() */
#include <linux/fs.h>		/* file_operations */
#include <linux/interrupt.h>	/* request_irq etc */
#include <linux/version.h>

MODULE_DESCRIPTION("pcidemo");
MODULE_AUTHOR("Stelian Pop/Pierre Ficheux, Open Wide");
MODULE_LICENSE("GPL");

/*
 * Supported devices
 */
#define VENDOR_ID	0x1969
#define DEVICE_ID	0x1062

static struct pci_device_id pcidemo_id_table[] __devinitdata = {
  {VENDOR_ID, DEVICE_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
  {0,}	/* 0 terminated list */
};
MODULE_DEVICE_TABLE(pci, pcidemo_id_table);

/*
 * Global variables
 */
static LIST_HEAD(pcidemo_list);

struct pcidemo_struct {
  struct list_head	link; /* Double linked list */
  struct pci_dev		*dev; /* PCI device */
  int			minor; /* Minor number */
  // Add mmio + mmio_len tabs
  /* ... */
};

/*
 * Event handlers
 */

/*
 * PCI handling
 */
static int __devinit pcidemo_probe(struct pci_dev *dev, const struct pci_device_id *ent)
{
  int ret = 0;
  struct pcidemo_struct *data;
  static int minor = 0;	

  printk(KERN_INFO "pcidemo: found %x:%x\n", ent->vendor, ent->device);

  /* Allocate a private structure and reference it as driver's data */
  data = (struct pcidemo_struct *)kmalloc(sizeof(struct pcidemo_struct), GFP_KERNEL);
  if (data == NULL) {
    printk(KERN_WARNING "pcidemo: unable to allocate private structure\n");

    ret = -ENOMEM;
    goto cleanup_kmalloc;
  }

  pci_set_drvdata(dev, data);

  /* Init private field */
  data->dev = dev;
  data->minor = minor++;

  /* Initialize device before it's used by the driver */
  ret = pci_enable_device(dev);
  if (ret < 0) {
    printk(KERN_WARNING "pcidemo: unable to initialize PCI device\n");

    goto cleanup_pci_enable;
  }

  /* Reserve PCI I/O and memory resources */
  ret = pci_request_regions(dev, "pcidemo");
  if (ret < 0) {
    printk(KERN_WARNING "pcidemo: unable to reserve PCI resources\n");
    goto cleanup_regions;
  }

  /* Inspect PCI BARs and remap IORESOURCE_MEM memory */
  /* ... */

  /* Install the irq handler */
  /* ... */

  /* Enable PCI bus-mastering for this device */
  pci_set_master(dev);

  /* Link the new data structure with others */
  list_add_tail(&data->link, &pcidemo_list);

  return 0;

 cleanup_irq:
  // iounmap regions

 cleanup_ioremap:
  pci_release_regions(dev);
 cleanup_regions:
  pci_disable_device(dev);
 cleanup_pci_enable:
  kfree(data);
 cleanup_kmalloc:
  return ret;
}

static void __devexit pcidemo_remove(struct pci_dev *dev)
{
  struct pcidemo_struct *data = pci_get_drvdata(dev);

  printk(KERN_INFO "pcidemo: device removed\n");

  // Free BARs => iounmap
  /* ... */

  pci_release_regions(dev);
  pci_disable_device(dev);

  // Free IRQ

  // remove from list
  list_del(&data->link);

  kfree(data);
}

static struct pci_driver pcidemo_driver = {
  .name =		"pcidemo",
  .id_table =	pcidemo_id_table,
  .probe =	pcidemo_probe,		/* Init one device */
  .remove =	pcidemo_remove,		/* Remove one device */
};

/*
 * Init and Exit
 */
static int __init pcidemo_init(void)
{
  int ret;

  /* Register PCI driver */
  ret = pci_register_driver(&pcidemo_driver);
  if (ret < 0) {
    printk(KERN_WARNING "pcidemo: unable to register PCI driver\n");
    return ret;
  }

  return 0;
}

static void __exit pcidemo_exit(void)
{
  pci_unregister_driver(&pcidemo_driver);
}

/*
 * Module entry points
 */
module_init(pcidemo_init);
module_exit(pcidemo_exit);
