/*
 * Includes
 */
#include <linux/kernel.h>	/* printk() */
#include <linux/module.h>	/* modules */
#include <linux/init.h>		/* module_{init,exit}() */
#include <linux/slab.h>		/* kmalloc()/kfree() */
#include <linux/pci.h>		/* pci_*() */
#include <linux/pci_ids.h>	/* pci idents */
#include <linux/list.h>		/* list_*() */
#include <asm/uaccess.h>	/* copy_{from,to}_user() */
#include <linux/fs.h>		/* file_operations */
#include <linux/interrupt.h>	/* request_irq etc */
#include <linux/version.h>

MODULE_DESCRIPTION("pcidemo");
MODULE_AUTHOR("Stelian Pop/Pierre Ficheux, Open Wide");
MODULE_LICENSE("GPL");

/*
 * Arguments
 */
static int major = 0; /* Major number */
module_param(major, int, 0660);
MODULE_PARM_DESC(major, "Static major number (none = dynamic)");

/*
 * Supported devices
 */

#define VENDOR_ID	0x1969
#define DEVICE_ID	0x1062

static struct pci_device_id pcidemo_id_table[] __devinitdata = {
  {VENDOR_ID, DEVICE_ID, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
  {0,}	/* 0 terminated list */
};
MODULE_DEVICE_TABLE(pci, pcidemo_id_table);

/*
 * Global variables
 */
static LIST_HEAD(pcidemo_list);

struct pcidemo_struct {
  struct list_head	link; /* Double linked list */
  struct pci_dev	*dev; /* PCI device */
  int	minor; /* Minor number */
  void	*mmio[DEVICE_COUNT_RESOURCE]; /* Remaped I/O memory */
  u32	mmio_len[DEVICE_COUNT_RESOURCE]; /* Size of remaped I/O memory */
};

/*
 * Event handlers
 */
irqreturn_t pcidemo_irq_handler(int irq, void *dev_id)
{
  struct pcidemo_struct *data = (struct pcidemo_struct *)dev_id;

  printk(KERN_DEBUG "pcidemo: interrupt from device %d\n", data->minor);
  return IRQ_NONE; // do nothing
}

/*
 * File operations
 */
static ssize_t pcidemo_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
  int i, real, bank = DEVICE_COUNT_RESOURCE;
  struct pcidemo_struct *data = file->private_data;

  /* Find the first remapped I/O memory bank to read */
  for (i = 0; i < DEVICE_COUNT_RESOURCE; i++) {
    if (data->mmio[i] != NULL) {
      bank = i;
      break;
    }
  }

  /* No bank found */
  if (bank == DEVICE_COUNT_RESOURCE) {
    printk(KERN_WARNING "pcidemo: no remaped I/O memory bank to read\n");
    return -ENXIO;
  }

  /* Check for overflow */
  real = min(data->mmio_len[bank] - (int)*ppos, count);

  // Send values to user-space 
  if (real) { 
    /* ... */
  }

  printk(KERN_DEBUG "pcidemo: read %d/%d chars at offset %d from remaped I/O memory bank %d\n", real, count, (int)*ppos, bank);

  /* Update ppos */
  /* ... */

  return real;
}

static int pcidemo_open(struct inode *inode, struct file *file)
{
  int minor = MINOR(inode->i_rdev);
  struct list_head *cur;
  struct pcidemo_struct *data;

  printk(KERN_DEBUG "pcidemo_open()\n");

  // Search in list for matching minor value
  list_for_each(cur, &pcidemo_list) {
    data = list_entry(cur, struct pcidemo_struct, link);

    if (data->minor == minor) {
      // Save data in VFS
      file->private_data = data;
      return 0;
    }
  }

  printk(KERN_WARNING "pcidemo: minor %d not found\n", minor);

  return -ENODEV;
}

static int pcidemo_release(struct inode *inode, struct file *file)
{
  printk(KERN_DEBUG "pcidemo_release()\n");

  file->private_data = NULL;

  return 0;
}

static struct file_operations pcidemo_fops = {
  .owner =	THIS_MODULE,
  .read =		pcidemo_read,
  .open =		pcidemo_open,
  .release =	pcidemo_release,
};

/*
 * PCI handling
 */
static int __devinit pcidemo_probe(struct pci_dev *dev, const struct pci_device_id *ent)
{
  int i, ret = 0;
  struct pcidemo_struct *data;
  static int minor = 0;
  u8 mypin;

  printk(KERN_INFO "pcidemo: found %x:%x\n", ent->vendor, ent->device);
  printk(KERN_INFO "pcidemo: using major %d and minor %d for this device\n", major, minor);

  /* Allocate a private structure and reference it as driver's data */
  data = (struct pcidemo_struct *)kmalloc(sizeof(struct pcidemo_struct), GFP_KERNEL);
  if (data == NULL) {
    printk(KERN_WARNING "pcidemo: unable to allocate private structure\n");
    ret = -ENOMEM;
    goto cleanup_kmalloc;
  }

  pci_set_drvdata(dev, data);

  /* Init private field */
  data->dev = dev;
  data->minor = minor++;

  /* Initialize device before it's used by the driver */
  ret = pci_enable_device(dev);
  if (ret < 0) {
    printk(KERN_WARNING "pcidemo: unable to initialize PCI device\n");
    goto cleanup_pci_enable;
  }

  /* Reserve PCI I/O and memory resources */
  ret = pci_request_regions(dev, "pcidemo");
  if (ret < 0) {
    printk(KERN_WARNING "pcidemo: unable to reserve PCI resources\n");
    goto cleanup_regions;
  }

  /* Inspect PCI BARs and remap I/O memory */
  for (i=0; i < DEVICE_COUNT_RESOURCE; i++) {
    if (pci_resource_len(dev, i) == 0)
      continue;

    if (pci_resource_start(dev, i) == 0)
      continue;

    printk(KERN_DEBUG "pcidemo: BAR %d (%#08x-%#08x), len=%d, flags=%#08x\n", i, (u32) pci_resource_start(dev, i), (u32) pci_resource_end(dev, i), (u32) pci_resource_len(dev, i), (u32) pci_resource_flags(dev, i));

    if (pci_resource_flags(dev, i) & IORESOURCE_MEM) {
      data->mmio[i] = ioremap(pci_resource_start(dev, i), pci_resource_len(dev, i));
      if (data->mmio[i] == NULL) {
	printk(KERN_WARNING "pcidemo: unable to remap I/O memory\n");

	ret = -ENOMEM;
	goto cleanup_ioremap;
      }

      data->mmio_len[i] = pci_resource_len(dev, i);

      printk(KERN_DEBUG "pcidemo: I/O memory has been remaped at %#08x\n", (u32)data->mmio[i]);
    } else {
      data->mmio[i] = NULL;
    }
  }

  /* Install the irq handler */
  pci_read_config_byte (dev, PCI_INTERRUPT_PIN, &mypin);
  if (mypin) {
    ret = request_irq(dev->irq, pcidemo_irq_handler, IRQF_SHARED, "pcidemo", data);
    if (ret < 0) {
      printk(KERN_WARNING "pcidemo: unable to register irq handler\n");

      goto cleanup_irq;
    }
  }
  else
    printk(KERN_WARNING "pcidemo: no IRQ!\n");

  /* Enable PCI bus-mastering for this device */
  pci_set_master(dev);

  /* Link the new data structure with others */
  list_add_tail(&data->link, &pcidemo_list);

  return 0;

 cleanup_irq:
  for (i--; i >= 0; i--) {
    if (data->mmio[i] != NULL)
      iounmap(data->mmio[i]);
  }
 cleanup_ioremap:
  pci_release_regions(dev);
 cleanup_regions:
  pci_disable_device(dev);
 cleanup_pci_enable:
  kfree(data);
 cleanup_kmalloc:
  return ret;
}

static void __devexit pcidemo_remove(struct pci_dev *dev)
{
  int i;
  u8 mypin;
  struct pcidemo_struct *data = pci_get_drvdata(dev);

  printk(KERN_INFO "pcidemo: device removed\n");

  for (i=0; i < DEVICE_COUNT_RESOURCE; i++) {
    if (data->mmio[i] != NULL)
      iounmap(data->mmio[i]);
  }

  pci_release_regions(dev);
  pci_disable_device(dev);

  pci_read_config_byte (dev, PCI_INTERRUPT_PIN, &mypin);
  if (mypin)
    free_irq(dev->irq, data);

  list_del(&data->link);
  kfree(data);
}

static struct pci_driver pcidemo_driver = {
  .name =		"pcidemo",
  .id_table =	pcidemo_id_table,
  .probe =	pcidemo_probe,		/* Init one device */
  .remove =	pcidemo_remove,		/* Remove one device */
};

/*
 * Init and Exit
 */
static int __init pcidemo_init(void)
{
  int ret;

  /* Register the char device driver */
  /* ... */

  /* Register PCI driver */
  ret = pci_register_driver(&pcidemo_driver);
  if (ret < 0) {
    printk(KERN_WARNING "pcidemo: unable to register PCI driver\n");
    unregister_chrdev(major, "pcidemo");

    return ret;
  }

  return 0;
}

static void __exit pcidemo_exit(void)
{
  pci_unregister_driver(&pcidemo_driver);

  // Unregister char driver
  /* ... */
}

/*
 * Module entry points
 */
module_init(pcidemo_init);
module_exit(pcidemo_exit);
