#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/proc_fs.h>	/* For /proc */

#define DRIVER_AUTHOR "Pierre Ficheux, pierre@ficheux.org"
#define DRIVER_DESC   "Panic Button driver: no URB + /proc"

#define VENDOR_ID	0x1130
#define PRODUCT_ID	0x0202

struct usb_panicb {
	struct usb_device *	udev;
	unsigned char		button;
};

/* Table of devices that work with this driver */
static struct usb_device_id id_table [] = {
	{ USB_DEVICE(VENDOR_ID, PRODUCT_ID) },
	{ },
};
MODULE_DEVICE_TABLE (usb, id_table);

/* Ask panic button for button status */
static int get_panicb_button_status (struct usb_panicb *panicb_dev)
{
  char *buf;
  int ret = 0;


  printk (KERN_INFO "get_panicb_button_status\n");

  // Allocate msg buffer
  if (!(buf = kmalloc(8, GFP_KERNEL))) {
    printk(KERN_WARNING "panicb: can't alloc buf\n");
    return -1;
  }

  memset (buf, 0, 8);	
  ret = usb_control_msg (panicb_dev->udev, usb_rcvctrlpipe (panicb_dev->udev, 0), 0x01, 0xA1, 0x300, 0x00, buf, 8, 2 * HZ);
  if (ret < 0) 
    printk (KERN_WARNING "panicb: IN, ret = %d\n", ret);
  else
    panicb_dev->button = *buf;
  
  kfree (buf);

  return ret;
}


/* Called as we read /proc/panicb */
static int panicb_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
  struct usb_interface *interface;
  struct usb_panicb *panicb_dev;
  int len = 0;

  if (off)
    return 0;

  printk (KERN_INFO "panicb_read_proc\n");

  interface = (struct usb_interface *)data;
  panicb_dev = usb_get_intfdata(interface);

  // Get button status
  get_panicb_button_status (panicb_dev);

  // Write it
  len += sprintf(page + len, "%d\n", panicb_dev->button);

  *eof = 1;

  return len;
}

static int panicb_probe (struct usb_interface *interface, const struct usb_device_id *id)
{
  struct usb_device *udev = interface_to_usbdev (interface);
  struct usb_panicb *panicb_dev;

  printk (KERN_INFO "panicb_probe: starting\n");

  panicb_dev = kmalloc (sizeof(struct usb_panicb), GFP_KERNEL);
  if (panicb_dev == NULL) {
    dev_err (&interface->dev, "Out of memory\n");
    return -ENOMEM;
  }

  // Fill private structure and save it
  memset (panicb_dev, 0x00, sizeof (*panicb_dev));
  panicb_dev->udev = usb_get_dev(udev);
  panicb_dev->button = 0;
  usb_set_intfdata (interface, panicb_dev);

  // Create /proc/panicb entry
  create_proc_read_entry("panicb", 0, NULL, panicb_read_proc, (void*)interface);

  dev_info(&interface->dev, "USB Panic Button device now attached\n");

  return 0;
}

static void panicb_disconnect(struct usb_interface *interface)
{
  struct usb_panicb *dev;

  dev = usb_get_intfdata (interface);
  usb_set_intfdata (interface, NULL);
  kfree(dev);

  remove_proc_entry ("panicb", NULL);

  dev_info(&interface->dev, "USB Panic Button now disconnected\n");
}

static struct usb_driver panicb_driver = {
	.name =		"panicb",
	.probe =	panicb_probe,
	.disconnect =	panicb_disconnect,
	.id_table =	id_table,
};

static int __init usb_panicb_init(void)
{
  int retval = 0;

  retval = usb_register(&panicb_driver);
  if (retval)
    err("usb_register failed. Error number %d", retval);

  return retval;
}

static void __exit usb_panicb_exit(void)
{
  usb_deregister(&panicb_driver);
}


module_init (usb_panicb_init);
module_exit (usb_panicb_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
