#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/usb.h>

#define DRIVER_AUTHOR "Pierre Ficheux, pierre@ficheux.org"
#define DRIVER_DESC "Panic Button driver: no URB + use /sys"

// Device definition
#define VENDOR_ID	0x1130
#define PRODUCT_ID	0x0202

struct usb_panicb {
	struct usb_device *	udev;
	unsigned int		button;
};

/* Table of devices that work with this driver */
static struct usb_device_id id_table [] = {
	{ USB_DEVICE(VENDOR_ID, PRODUCT_ID) },
	{ },
};
MODULE_DEVICE_TABLE (usb, id_table);

/* Ask panic button for button status (0/1) */
static int get_panicb_button_status (struct usb_panicb *panicb_dev)
{
  char *buf;
  int ret = 0;

  // Allocate msg buffer
  if (!(buf = kmalloc(8, GFP_KERNEL))) {
    printk(KERN_WARNING "panicb: can't alloc buf\n");
    return -1;
  }

  memset (buf, 0, 8);	
  ret = usb_control_msg (panicb_dev->udev, usb_rcvctrlpipe (panicb_dev->udev, 0), 0x01, 0xA1, 0x300, 0x00, buf, 8, 2 * HZ);
  if (ret < 0) 
    printk (KERN_WARNING "panicb: IN, ret = %d\n", ret);
  else
    panicb_dev->button = *buf;
  
  kfree (buf);

  return ret;
}


static ssize_t show_button(struct device *dev, struct device_attribute *attr, char *buf)
{
  struct usb_interface *intf = to_usb_interface(dev); 
  struct usb_panicb *panicb_dev = usb_get_intfdata(intf);

  get_panicb_button_status (panicb_dev);

  return sprintf(buf, "%d\n", panicb_dev->button); 
}


static ssize_t set_button(struct device *dev, struct device_attribute *attr, const char *buf, size_t count)
{
	return count;
}
static DEVICE_ATTR(button, S_IWUGO | S_IRUGO, show_button, set_button);



static int panicb_probe (struct usb_interface *interface, const struct usb_device_id *id)
{
  struct usb_device *udev = interface_to_usbdev (interface);
  struct usb_panicb *panicb_dev;

  panicb_dev = kmalloc (sizeof(struct usb_panicb), GFP_KERNEL);
  if (panicb_dev == NULL) {
    dev_err (&interface->dev, "Out of memory\n");
    return -ENOMEM;
  }

  // Fill private structure and save it
  memset (panicb_dev, 0x00, sizeof (*panicb_dev));
  panicb_dev->udev = usb_get_dev(udev);
  panicb_dev->button = 0;
  usb_set_intfdata (interface, panicb_dev);

  device_create_file(&interface->dev, &dev_attr_button);

  dev_info(&interface->dev, "USB Panic Button device now attached\n");


  return 0;
}

static void panicb_disconnect(struct usb_interface *interface)
{
  struct usb_panicb *dev;

  dev = usb_get_intfdata (interface);
  usb_set_intfdata (interface, NULL);

  device_remove_file(&interface->dev, &dev_attr_button);
  usb_put_dev(dev->udev);

  kfree(dev);

  dev_info(&interface->dev, "USB Panic Button now disconnected\n");
}

static struct usb_driver panicb_driver = {
	.name =		"panicb",
	.probe =	panicb_probe,
	.disconnect =	panicb_disconnect,
	.id_table =	id_table,
};

static int __init usb_panicb_init(void)
{
  int retval = 0;

  retval = usb_register(&panicb_driver);
  if (retval)
    err("usb_register failed. Error number %d", retval);

  return retval;
}

static void __exit usb_panicb_exit(void)
{
  usb_deregister(&panicb_driver);
}


module_init (usb_panicb_init);
module_exit (usb_panicb_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
